import unittest
import shutil
import filecmp
import copy
import os
import sys
import pyfastaq
import pymummer
from circlator import start_fixer

modules_dir = os.path.dirname(os.path.abspath(start_fixer.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestStartFixer(unittest.TestCase):
    def test_init(self):
        '''Test init'''
        contigs_fa = os.path.join(data_dir, 'start_fixer_init_contigs.fa')
        dna_da = os.path.join(data_dir, 'start_fixer_init_dnaA.fa')

        with self.assertRaises(start_fixer.Error):
            sfixer = start_fixer.StartFixer('notafile', 'outprefix')

        sfixer = start_fixer.StartFixer(contigs_fa, 'outprefix')
        expected_contigs = {
            'contig1': pyfastaq.sequences.Fasta('contig1', 'ACGT'),
            'contig2': pyfastaq.sequences.Fasta('contig2', 'AAAA'),
        }
        self.assertEqual(expected_contigs, sfixer.input_assembly)
        self.assertEqual(set(), sfixer.ignore)

        with self.assertRaises(start_fixer.Error):
            sfixer = start_fixer.StartFixer(contigs_fa, 'outprefix', genes_fa='notafile')

        ignore_file = os.path.join(data_dir, 'start_fixer_init_ignore_ids')
        sfixer = start_fixer.StartFixer(contigs_fa, 'outprefix', ignore=ignore_file)
        self.assertEqual({'ignore_me1', 'ignore_me2'}, sfixer.ignore)


    def test_rename_contigs(self):
        '''Test _rename_contigs'''
        contigs_in = {
            'ctg1': pyfastaq.sequences.Fasta('ctg1', 'ACGT'),
            'ctg2 foo bar': pyfastaq.sequences.Fasta('ctg2 foo bar', 'AAA'),
        }
        expected_contigs = {
            'ctg1': pyfastaq.sequences.Fasta('ctg1', 'ACGT'),
            'ctg2': pyfastaq.sequences.Fasta('ctg2', 'AAA'),
        }
        expected_names = {'ctg1': 'ctg1', 'ctg2': 'ctg2 foo bar'}
        got_contigs, got_names = start_fixer.StartFixer._rename_contigs(contigs_in)
        self.assertEqual(expected_names, got_names)
        self.assertEqual(expected_contigs, got_contigs)
        contigs_in['ctg2 abc'] = pyfastaq.sequences.Fasta('ctg2 abc', 'AAA')
        with self.assertRaises(start_fixer.Error):
            start_fixer.StartFixer._rename_contigs(contigs_in)


    def test_write_renamed_contigs(self):
        '''Test _write_renamed_contigs'''
        contigs_dict = {
            'ctg1': pyfastaq.sequences.Fasta('ctg1', 'ACGT'),
            'ctg2': pyfastaq.sequences.Fasta('ctg2', 'AAA'),
        }
        rename_dict = {'ctg1': 'ctg1', 'ctg2': 'ctg2 foo bar'}
        tmp_out = 'tmp.test_write_renamed_contigs.fa'
        start_fixer.StartFixer._write_renamed_contigs(contigs_dict, rename_dict, tmp_out)
        expected = os.path.join(data_dir, 'start_fixer_write_renamed_contigs.fa')
        self.assertTrue(filecmp.cmp(expected, tmp_out, shallow=False))
        os.unlink(tmp_out)


    def test_max_length_from_fasta_file(self):
        '''Test _max_length_from_fasta_file'''
        infile = os.path.join(data_dir, 'start_fixer_max_length_from_fasta_file.fa')
        self.assertEqual(11, start_fixer.StartFixer._max_length_from_fasta_file(infile))


    def test_write_fasta_plus_circularized_ends(self):
        '''Test _write_fasta_plus_circularized_ends'''
        infile = os.path.join(data_dir, 'start_fixer_write_fasta_plus_circularized_ends.in.fa')
        expected = os.path.join(data_dir, 'start_fixer_write_fasta_plus_circularized_ends.out.fa')
        expected_ignore = os.path.join(data_dir, 'start_fixer_write_fasta_plus_circularized_ends.out.ignore.fa')
        tmp_out = 'tmp.test_write_fasta_plus_circularized_ends.fa'
        contigs = {}
        pyfastaq.tasks.file_to_dict(infile, contigs)
        got = start_fixer.StartFixer._write_fasta_plus_circularized_ends(contigs, tmp_out, 5)
        self.assertEqual(6, got)
        self.assertTrue(filecmp.cmp(expected, tmp_out, shallow=False))
        got = start_fixer.StartFixer._write_fasta_plus_circularized_ends(contigs, tmp_out, 5, ignore={'seq1', 'seq4'})
        self.assertEqual(3, got)
        self.assertTrue(filecmp.cmp(expected_ignore, tmp_out, shallow=False))
        os.unlink(tmp_out)
        got = start_fixer.StartFixer._write_fasta_plus_circularized_ends(contigs, tmp_out, 5, ignore={'seq1', 'seq2', 'seq3', 'seq4'})
        self.assertEqual(0, got)
        self.assertFalse(os.path.exists(tmp_out))


    def test_find_circular_using_promer(self):
        '''Test _find_circular_using_promer'''
        contigs_infile = os.path.join(data_dir, 'start_fixer_find_circular_using_promer.contigs.fa')
        ref_genes_fa = os.path.join(data_dir, 'start_fixer_find_circular_using_promer.refs.fa')
        tmp_outprefix = 'tmp.start_fixer_find_circular_using_promer'
        contigs_dict = {}
        pyfastaq.tasks.file_to_dict(contigs_infile, contigs_dict)
        end_extend = start_fixer.StartFixer._max_length_from_fasta_file(ref_genes_fa)

        expected = {
            'ctg_DNAA_ECOLI': pymummer.alignment.Alignment('1021\t2424\t1\t1404\t1404\t1404\t100.00\t.\t.\t3404\t1404\t1\t1\tctg_DNAA_ECOLI\tsp|P03004|DNAA_ECOLI\n'),
            'ctg_DNAA1_CHLPN': pymummer.alignment.Alignment('685\t2067\t1\t1383\t1383\t1383\t100.00\t.\t.\t2808\t1383\t1\t1\tctg_DNAA1_CHLPN__ends\tsp|Q9Z8M9|DNAA1_CHLPN\n'),
            'ctg_DNAA_ECOLI_2': pymummer.alignment.Alignment('901\t2304\t1\t1404\t1404\t1404\t100.00\t.\t.\t3044\t1404\t1\t1\tctg_DNAA_ECOLI_2\tsp|P03004|DNAA_ECOLI_2\n'),
}

        got = start_fixer.StartFixer._find_circular_using_promer(tmp_outprefix, ref_genes_fa, contigs_dict, 70, end_extend, sys.stdout)
        self.assertEqual(expected, got)
        os.unlink(tmp_outprefix + '.contigs_with_ends.fa')
        os.unlink(tmp_outprefix + '.promer')
        got = start_fixer.StartFixer._find_circular_using_promer(tmp_outprefix, ref_genes_fa, contigs_dict, 70, end_extend, sys.stdout, ignore={x for x in contigs_dict})
        self.assertEqual({}, got)


    def test_find_circular_using_prodigal(self):
        '''test _find_circular_using_prodigal'''
        contigs_infile = os.path.join(data_dir, 'start_fixer_find_circular_using_prodigal.ctgs.fa')
        outprefix = 'tmp.test_find_circular_using_prodigal'
        contigs_dict = {}
        pyfastaq.tasks.file_to_dict(contigs_infile, contigs_dict)
        circular_from_promer = {'ctg3': 'foo'}
        got = start_fixer.StartFixer._find_circular_using_prodigal(outprefix, contigs_dict, circular_from_promer, sys.stdout)
        # We should have a dictionary with one key == "ctg1" and a value like this:
        # 'ctg1\tProdigal_v2.60\tCDS\t841\t2244\t103.8\t+\t0\tID=1_1;partial=00;start_type=ATG;rbs_motif=TAAAA;rbs_spacer=4bp;gc_cont=0.486;conf=100.00;score=103.84;cscore=95.60;sscore=8.24;rscore=4.24;uscore=-0.13;tscore=4.12;'
        # ... but this depends on the version of prodigal, so is not predictable
        self.assertEqual({'ctg1'}, set(got.keys()))
        got_fields = got['ctg1'].split('\t')
        self.assertEqual(got_fields[0], 'ctg1')
        self.assertTrue(got_fields[1].startswith('Prodigal'))
        os.unlink(outprefix + '.for_prodigal.fa')
        os.unlink(outprefix + '.prodigal.gff')
        got = start_fixer.StartFixer._find_circular_using_prodigal(outprefix, contigs_dict, circular_from_promer, sys.stdout, ignore={'ctg1', 'ctg2'})
        self.assertEqual({}, got)


    def test_rearrange_contigs(self):
        '''test _rearrange_contigs'''
        # try all the combinations of promer and prodigal and ignore. Also whether or not to
        # reverse complement and DNAA spanning across start/end of contig
        # promer_ctg1: has sp|P03004|DNAA_ECOLI starting at position 2000 on fwd strand
        # promer_ctg2: has sp|P03004|DNAA_ECOLI_2 starting at position 3403 on reverse strand (=revcomp of promer_ctg1)
        # promer_ctg3: has sp|P03004|DNAA_ECOLI starts at 4745, split over end/start of contig
        # promer_ctg4: has sp|P03004|DNAA_ECOLI starts at 644, split over start/end of contig (=revcomp of promer_ctg3)
        # ignore_ctg: same as promer_ctg1, but in the ignore set, so should be unchanged
        # prodigal_ctg1: has a two genes, nearest to middle starts at 4529
        # prodigal_ctg2: same as ctg1, but reverse complement, gene starts at 4862 on rev strand
        contigs_infile = os.path.join(data_dir, 'start_fixer_rearrange_contigs.in.fa')
        ref_genes_fa = os.path.join(data_dir, 'start_fixer_rearrange_contigs.refs.fa')
        tmp_outprefix = 'tmp.test_rearrange_contigs'
        tmp_log = tmp_outprefix + '.log'
        contigs_dict = {}
        pyfastaq.tasks.file_to_dict(contigs_infile, contigs_dict)
        to_ignore = {'ignore_ctg'}
        end_extend = start_fixer.StartFixer._max_length_from_fasta_file(ref_genes_fa)
        circ_with_promer = start_fixer.StartFixer._find_circular_using_promer(tmp_outprefix, ref_genes_fa, contigs_dict, 70, end_extend, sys.stdout, ignore=to_ignore)
        circ_with_prodigal = start_fixer.StartFixer._find_circular_using_prodigal(tmp_outprefix, contigs_dict, circ_with_promer, sys.stdout, ignore=to_ignore)
        start_fixer.StartFixer._rearrange_contigs(contigs_dict, circ_with_promer, circ_with_prodigal, to_ignore, end_extend, tmp_log)
        expected_log = os.path.join(data_dir, 'start_fixer_rearrange_contigs.expect.log')
        self.assertTrue(filecmp.cmp(expected_log, tmp_log, shallow=False))
        expected_dict = {}
        expected_fa = os.path.join(data_dir, 'start_fixer_rearrange_contigs.expect.fa')
        pyfastaq.tasks.file_to_dict(expected_fa, expected_dict)
        self.assertEqual(expected_dict, contigs_dict)
        os.unlink(tmp_outprefix + '.contigs_with_ends.fa')
        os.unlink(tmp_outprefix + '.for_prodigal.fa')
        os.unlink(tmp_outprefix + '.prodigal.gff')
        os.unlink(tmp_outprefix + '.promer')
        os.unlink(tmp_log)


    def test_run_ignore_all(self):
        '''test run when ignoring all sequences'''
        tmp_prefix = 'tmp.start_fixer.test_run_when_ignoring_all'
        input_fa = os.path.join(data_dir, 'start_fixer_run_ignore_all.fa')
        to_ignore = os.path.join(data_dir, 'start_fixer_run_ignore_all.to_ignore')
        sfixer = start_fixer.StartFixer(input_fa, tmp_prefix, ignore=to_ignore)
        sfixer.run()
        expected_fa = os.path.join(data_dir, 'start_fixer_run_ignore_all.expect.fa')
        self.assertTrue(filecmp.cmp(expected_fa, tmp_prefix + '.fasta'))
        for suffix in ['detailed.log', 'fasta', 'log', 'promer.contigs_with_ends.fa', 'promer.promer']:
            try:
                os.unlink(tmp_prefix + '.' + suffix)
            except:
                pass


    def test_run_none_for_prodigal(self):
        '''test run when everthing sorted by promer'''
        tmp_prefix = 'tmp.start_fixer.test_run_none_for_prodigal'
        input_ctg = os.path.join(data_dir, 'start_fixer_run_none_for_prodigal.ctg.fa')
        input_ref = os.path.join(data_dir, 'start_fixer_run_none_for_prodigal.ref.fa')
        sfixer = start_fixer.StartFixer(input_ctg, tmp_prefix, genes_fa=input_ref)
        sfixer.run()
        expected_fa = os.path.join(data_dir, 'start_fixer_run_none_for_prodigal.expect.fa')
        self.assertTrue(filecmp.cmp(expected_fa, tmp_prefix + '.fasta'))
        for suffix in ['detailed.log', 'fasta', 'log', 'promer.contigs_with_ends.fa', 'promer.promer']:
            try:
                os.unlink(tmp_prefix + '.' + suffix)
            except:
                pass


    def test_run_bit_of_everything(self):
        '''test run with promer, prodigal, and ignoring a sequence'''
        tmp_prefix = 'tmp.start_fixer.test_run_bit_of_everything'
        input_ctg = os.path.join(data_dir, 'start_fixer_run_bit_of_everything.ctg.fa')
        input_ref = os.path.join(data_dir, 'start_fixer_run_bit_of_everything.ref.fa')
        input_ignore = os.path.join(data_dir, 'start_fixer_run_bit_of_everything.ignore')
        sfixer = start_fixer.StartFixer(input_ctg, tmp_prefix, genes_fa=input_ref, ignore=input_ignore)
        sfixer.run()
        expected_fa = os.path.join(data_dir, 'start_fixer_run_bit_of_everything.expect.fa')
        self.assertTrue(filecmp.cmp(expected_fa, tmp_prefix + '.fasta'))

        for suffix in ['detailed.log', 'fasta', 'log', 'promer.contigs_with_ends.fa', 'promer.promer', 'prodigal.for_prodigal.fa', 'prodigal.prodigal.gff']:
            try:
                os.unlink(tmp_prefix + '.' + suffix)
            except:
                pass



import os
import sys
import shutil
import tempfile
import pyfastaq
import logging
from circlator import common

class Error (Exception): pass

class Merger:
    def __init__(self,
      assembly_fasta,
      outprefix,
      merge=True,
    ):
        self.input_assembly = os.path.abspath(assembly_fasta)
        self.outprefix = os.path.abspath(outprefix)
        self.run_pre_merge = merge

        if not os.path.exists(self.input_assembly):
            raise Error('Input assembly "' + self.input_assembly + '" not found. Cannot continue')

        # need to have toAmos and minimus2 in user's path. We are
        # not using external_progs.py for this because toAmos and minimus2
        # don't output version information. We'll just check we can find
        # them and assume the version is OK
        for program in ['toAmos', 'minimus2']:
            self.check_path(program)

        self.log_file = self.outprefix + '.log'


    def check_path(self, program):
        if shutil.which(program) is None:
            print(program + ' not found in path! Cannot continue')
            raise Error(program + ' not found in path! Cannot continue')


    def _write_split_contig(self, contig, outfile):
        half_coord = int(len(contig) / 2)
        seq1 = contig.subseq(0, half_coord)
        seq2 = contig.subseq(half_coord, len(contig))
        seq1.id += '.1'
        seq2.id += '.2'
        f = pyfastaq.utils.open_file_write(outfile)
        print(seq1, file=f)
        print(seq2, file=f)
        pyfastaq.utils.close(f)


    def _run_minimus2(self, infile, outdir):
        amos_afg_prefix = os.path.join(outdir, 'minimus2')
        amos_afg = amos_afg_prefix + '.afg'
        cmd = 'toAmos -s ' + infile + ' -o ' + amos_afg
        common.syscall(cmd)
        cmd = 'minimus2 ' + amos_afg_prefix
        return common.syscall(cmd, allow_fail=True)


    def _run_minimus2_on_one_contig(self, contig):
        tmpdir = tempfile.mkdtemp(prefix=self.outprefix + 'tmp.minimus2.', dir=os.getcwd())
        split_fa = os.path.join(tmpdir, 'in.fasta')
        self._write_split_contig(contig, split_fa)
        finished_ok, errors = self._run_minimus2(split_fa, tmpdir)

        if not finished_ok:
            logging.info('Minimus2 returned non-zero error code. Keeping original contig ' + contig.id)
            shutil.rmtree(tmpdir)
            return contig, False

        singletons = {}
        circularised = {}
        pyfastaq.tasks.file_to_dict(os.path.join(tmpdir, 'minimus2.singletons.seq'), singletons)
        pyfastaq.tasks.file_to_dict(os.path.join(tmpdir, 'minimus2.fasta'), circularised)
        contig_was_circularised = False

        if len(singletons) == 0 and len(circularised) == 1:
            new_contig = list(circularised.values())[0]
            new_contig.id = contig.id + '.circularised'
            contig_was_circularised = True
        elif len(singletons) == 2:
            new_contig = contig
        else:
            print('Error running minimus2 on contig', contig.id, '- there were', len(singletons), 'singletons and', len(circularised), 'circularised contigs. Cannot continue', file=sys.stderr)
            print('minimus2 was run in this directory:', tmpdir, file=sys.stderr)
            raise Error('Error running minimus2. Cannot continue')

        shutil.rmtree(tmpdir)
        return new_contig, contig_was_circularised


    def _run_minimus2_on_each_contig(self, infile, outfile):
        seq_reader = pyfastaq.sequences.file_reader(infile)
        f_out = pyfastaq.utils.open_file_write(outfile)
        circularised_contigs = []

        for contig in seq_reader:
            logging.info('Circularising: ' + contig.id + ' ...')
            new_contig, circularised = self._run_minimus2_on_one_contig(contig)
            print(new_contig, file=f_out)
            if circularised:
                circularised_contigs.append(contig.id)
            logging.info(contig.id + ' circularised: ' + str(circularised))

        pyfastaq.utils.close(f_out)
        logging.info('Circularised contigs:\t' + '\t'.join([x for x in circularised_contigs]))


    def _run_initial_merge(self, infile, outdir):
        try:
            os.mkdir(outdir)
        except:
            raise Error('Error making directory ' + outdir)

        original_contigs = {}
        pyfastaq.tasks.file_to_dict(infile, original_contigs)
        finished_ok, errors = self._run_minimus2(infile, outdir)
        logging.info('First minimus2 run OK: ' +  str(finished_ok))

        if finished_ok:
            merge_fasta = self.outprefix + '.merge.fasta'
            merged_contigs = {}
            singleton_contigs = {}
            pyfastaq.tasks.file_to_dict(os.path.join(outdir, 'minimus2.fasta'), merged_contigs)
            pyfastaq.tasks.file_to_dict(os.path.join(outdir, 'minimus2.singletons.seq'), singleton_contigs)

            contigs_that_were_merged = [x for x in original_contigs if x not in singleton_contigs]
            logging.info('Number of original contigs: ' + str(len(original_contigs)))
            logging.info('Number of singleton contigs: ' + str(len(singleton_contigs)))
            logging.info('Number of contigs that were merged: ' +  str(len(contigs_that_were_merged)))
            logging.info('Merged contigs: ' + '\t'.join(merged_contigs))

            f_out = pyfastaq.utils.open_file_write(merge_fasta)
            for name, contig in merged_contigs.items():
                print(contig, file=f_out)
            for name, contig in singleton_contigs.items():
                print(contig, file=f_out)
            pyfastaq.utils.close(f_out)

            logging.info('Number of contigs to try to circularise: ' + str(len(merged_contigs) + len(singleton_contigs)))
            return merge_fasta
        else:
            print('Error running merge. Going to use original contigs. Minimus2 errors follow...\n', file=sys.stderr)
            print(errors, file=sys.stderr)
            print('... end of Minimus2 output.\n\n', file=sys.stderr)
            logging.info('Number of original contigs: ' + str(len(original_contigs)))
            logging.info('Number of singleton contigs: NA')
            logging.info('Number of contigs that were merged: NA')
            logging.info('Merged contigs: NA')
            return infile


    def run(self):
        logging.basicConfig(
            filename=self.log_file,
            format='%(message)s',
            filemode='w',
            level=logging.INFO
        )

        if self.run_pre_merge:
            merge_dir = self.outprefix + '.merge'
            merge_fasta = self._run_initial_merge(self.input_assembly, merge_dir)
        else:
            merge_fasta = self.input_assembly

        final_fasta = self.outprefix + '.circularise.fasta'
        self._run_minimus2_on_each_contig(merge_fasta, final_fasta)




import argparse
import sys
import pyfastaq
import circlator

def run():
    parser = argparse.ArgumentParser(
        description = 'Make reads from mapping to be reassembled',
        usage = 'circlator bam2reads [options] <in.bam> <outprefix>')
    parser.add_argument('--discard_unmapped', action='store_true', help='Use this to not keep unmapped reads')
    parser.add_argument('--fastq', action='store_true', help='Write fastq output (if quality scores are present in input BAM file)')
    parser.add_argument('--only_contigs', help='File of contig names (one per line). Only reads that map to these contigs are kept (and unmapped reads, unless --discard_unmapped is used).', metavar='FILENAME')
    parser.add_argument('--length_cutoff', type=int, help='All reads mapped to contigs shorter than this will be kept [%(default)s]', default=100000, metavar='INT')
    parser.add_argument('--min_read_length', type=int, help='Minimum length of read to output [%(default)s]', default=250, metavar='INT')
    parser.add_argument('--split_all_reads', action='store_true', help='By default, reads mapped to shorter contigs are left unchanged. This option splits them into two, broken at the middle of the contig to try to force circularization. May help if the assembler does not detect circular contigs (eg canu)')
    parser.add_argument('--verbose', action='store_true', help='Be verbose')
    parser.add_argument('bam', help='Name of input bam file', metavar='in.bam')
    parser.add_argument('outprefix', help='Prefix of output filenames')
    options = parser.parse_args()

    bam_filter = circlator.bamfilter.BamFilter(
        options.bam,
        options.outprefix,
        fastq_out=options.fastq,
        length_cutoff=options.length_cutoff,
        min_read_length=options.min_read_length,
        contigs_to_use=options.only_contigs,
        discard_unmapped=options.discard_unmapped,
        split_all_reads=options.split_all_reads,
        verbose=options.verbose,
    )
    bam_filter.run()




import unittest
import shutil
import filecmp
import copy
import os
import pyfastaq
from circlator import assembly

modules_dir = os.path.dirname(os.path.abspath(assembly.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestAssembly(unittest.TestCase):
    def test_init_file_not_dir(self):
        '''Test _init_file_not_dir'''
        contigs_fasta = os.path.join(data_dir, 'assembly_test_init_contigs.fasta')
        a = assembly.Assembly(contigs_fasta, 'spades')
        self.assertEqual(a.contigs_fasta, contigs_fasta)
        self.assertIsNone(a.contigs_fastg)
        self.assertIsNone(a.assembly_graph_fastg)
        self.assertIsNone(a.assembler_dir)


    def test_init_pre_3_6_1_ok(self):
        '''Test _init_pre_3_6_1_ok'''
        test_dir = os.path.join(data_dir, 'assembly_test_init_spades_pre_3_6_1_ok')
        a = assembly.Assembly(test_dir, 'spades')
        self.assertEqual(a.contigs_fasta, os.path.join(test_dir, 'contigs.fasta'))
        self.assertEqual(a.contigs_fastg, os.path.join(test_dir, 'contigs.fastg'))
        self.assertEqual(a.assembler_dir, test_dir)
        self.assertIsNone(a.assembly_graph_fastg)
        self.assertIsNone(a.contigs_paths)


    def test_init_post_3_6_1_ok(self):
        '''Test _init_post_3_6_1_ok'''
        test_dir = os.path.join(data_dir, 'assembly_test_init_spades_post_3_6_1_ok')
        a = assembly.Assembly(test_dir, 'spades')
        self.assertEqual(a.assembler_dir, test_dir)
        self.assertEqual(a.contigs_fasta, os.path.join(test_dir, 'contigs.fasta'))
        self.assertEqual(a.assembly_graph_fastg, os.path.join(test_dir, 'assembly_graph.fastg'))
        self.assertEqual(a.contigs_paths, os.path.join(test_dir, 'contigs.paths'))
        self.assertIsNone(a.contigs_fastg)


    def test_init_raises_error(self):
        '''Test _init_raises_error'''
        test_dirs = [
            'assembly_test_init_no_fasta',
            'assembly_test_init_spades_pre_3_6_1_no_fastg',
            'assembly_test_init_spades_post_3_6_1_no_contigs_paths',
            'assembly_test_init_spades_post_3_6_1_no_assembly_graph',
            'assembly_test_init_spades_post_3_6_1_no_assembly_graph_or_contig_paths',
        ]

        for test_dir in test_dirs:
            with self.assertRaises(assembly.Error):
                a = assembly.Assembly(os.path.join(data_dir, test_dir), 'spades')


    def test_get_contigs(self):
        '''Test get_contigs'''
        a = assembly.Assembly(os.path.join(data_dir, 'assembly_test_get_contigs.fasta'), 'spades')
        expected = {
            'contig1': pyfastaq.sequences.Fasta('contig1', 'ACGT'),
            'contig2': pyfastaq.sequences.Fasta('contig2', 'AAAA'),
        }
        self.assertEqual(expected, a.get_contigs())


    def test_circular_contigs_from_spades_before_3_6_1(self):
        '''Test _circular_contigs_from_spades_before_3_6_1'''
        fastg_file = os.path.join(data_dir, 'assembly_test_circular_contigs_from_spades_before_3_6_1.fastg')
        got = assembly.Assembly._circular_contigs_from_spades_before_3_6_1(fastg_file)
        expected = {'NODE_1_length_5_cov_42.42_ID_1'}
        self.assertEqual(expected, got)


    def test_spades_contigs_paths_to_dict(self):
        '''Test _spades_contigs_paths_to_dict'''
        contigs_paths = os.path.join(data_dir, 'assembly_test_spades_contigs_paths_to_dict.contigs.paths')
        expected = {
            "NODE_1_length_86545_cov_16.2519_ID_2261": "1-,2+",
            "NODE_1_length_86545_cov_16.2519_ID_2261'": "2-,1+",
            "NODE_2_length_22106_cov_99.9797_ID_2263": "4+,10-,11+,10+,9-,5-,3+,5-",
            "NODE_2_length_22106_cov_99.9797_ID_2263'": "5+,3-,5+,9+,10-,11-,10+,4-",
            "NODE_3_length_21556_cov_27.4797_ID_2265": "6+",
            "NODE_3_length_21556_cov_27.4797_ID_2265'": "6-",
            "NODE_4_length_16761_cov_36.4471_ID_2267": "7+",
            "NODE_4_length_16761_cov_36.4471_ID_2267'": "7-",
            "NODE_5_length_8134_cov_20.6513_ID_2269": "14+",
            "NODE_5_length_8134_cov_20.6513_ID_2269'": "14-",
            "NODE_6_length_6712_cov_6.68474_ID_2271": "15+",
            "NODE_6_length_6712_cov_6.68474_ID_2271'": "15-",
            "NODE_7_length_4806_cov_3.33747_ID_2273": "13+",
            "NODE_7_length_4806_cov_3.33747_ID_2273'": "13-",
            "NODE_8_length_4566_cov_19.8709_ID_2275": "12+",
            "NODE_8_length_4566_cov_19.8709_ID_2275'": "12-",
            "NODE_9_length_3068_cov_67.4016_ID_2277": "8+",
            "NODE_9_length_3068_cov_67.4016_ID_2277'": "8-",
        }
        got = assembly.Assembly._spades_contigs_paths_to_dict(contigs_paths)
        self.assertEqual(expected, got)


    def test_circular_edges_to_edge_numbers_dict(self):
        '''Test _circular_edges_to_edge_numbers_dict'''
        edges = {'NODE_1_length_86545_cov_16.2519_ID_2261', 'NODE_5_length_8134_cov_20.6513_ID_2269'}
        expected = {
            '1': 'NODE_1_length_86545_cov_16.2519_ID_2261',
            '5': 'NODE_5_length_8134_cov_20.6513_ID_2269'
        }
        got = assembly.Assembly._circular_edges_to_edge_numbers_dict(edges)
        self.assertEqual(expected, got)


    def test_circular_contigs_from_spades_after_3_6_1(self):
        '''Test _circular_contigs_from_spades_after_3_6_1'''
        contigs_paths = os.path.join(data_dir, 'assembly_test_circular_contigs_from_spades_after_3_6_1.contigs.paths')
        assembly_graph_fastg = os.path.join(data_dir, 'assembly_test_circular_contigs_from_spades_after_3_6_1.assembly_graph.fastg')
        got = assembly.Assembly._circular_contigs_from_spades_after_3_6_1(assembly_graph_fastg, contigs_paths)
        expected = {
            'NODE_5_length_8134_cov_20.6513_ID_2269',
            'NODE_6_length_6712_cov_6.68474_ID_2271',
            'NODE_7_length_4806_cov_3.33747_ID_2273',
            'NODE_8_length_4566_cov_19.8709_ID_2275',
        }
        self.assertEqual(expected, got)


    def test_circular_contigs_from_canu_gfa(self):
        '''Test _circular_contigs_from_canu_gfa'''
        infile = os.path.join(data_dir, 'assembly_test_circular_contigs_from_canu.gfa')
        got = assembly.Assembly._circular_contigs_from_canu_gfa(infile)
        expected = {'tig00000042'}
        self.assertEqual(expected, got)


    def test_circular_contigs_spades_pre_3_6_1(self):
        '''Test circular_contigs with spades pre 3.6.1'''
        a = assembly.Assembly(os.path.join(data_dir, 'assembly_test_circular_contigs_spades_pre_3_6_1'), 'spades')
        got = a.circular_contigs()
        expected = {'NODE_1_length_5_cov_42.42_ID_1'}
        self.assertEqual(expected, got)


    def test_circular_contigs_spades_post_3_6_1(self):
        '''Test circular_contigs with spades post 3.6.1'''
        a = assembly.Assembly(os.path.join(data_dir, 'assembly_test_circular_contigs_spades_post_3_6_1'), 'spades')
        got = a.circular_contigs()
        expected = {
            'NODE_5_length_8134_cov_20.6513_ID_2269',
            'NODE_6_length_6712_cov_6.68474_ID_2271',
            'NODE_7_length_4806_cov_3.33747_ID_2273',
            'NODE_8_length_4566_cov_19.8709_ID_2275',
        }
        self.assertEqual(expected, got)


    def test_circular_contigs_just_fasta(self):
        '''Test circular_contigs when input is just fasta'''
        a = assembly.Assembly(os.path.join(data_dir, 'assembly_test_circular_contigs_only_contigs.fasta'), 'spades')
        got = a.circular_contigs()
        expected = set()
        self.assertEqual(expected, got)



import os
import pyfastaq

class Error (Exception): pass

class Assembly:
    def __init__(self, path, assembler):
        '''path can be a directory or a filename. If directory, assumes the name of a SPAdes
           or Canu (with files contigs.fasta and contigs.gfa files)
           output directory. If a file, assumes it is a fasta file of contigs'''
        self.assembler = assembler

        if not os.path.exists(path):
            raise Error('Input path to Assembly.__init__ not found: ' + path)
        elif os.path.isdir(path):
            self.assembler_dir = os.path.abspath(path)
        else:
            self.contigs_fasta = os.path.abspath(path)
            self.assembler_dir = None

        self._set_filenames()


    @staticmethod
    def _file_exists(filename):
        if os.path.exists(filename):
            return filename
        else:
            return None


    def _set_filenames(self):
        self.contigs_gfa = None
        self.contigs_fastg = None
        self.contigs_paths = None
        self.assembly_graph_fastg = None

        if self.assembler_dir is None:
            return

        contigs_fasta = os.path.join(self.assembler_dir, 'contigs.fasta')
        self.contigs_fasta = self._file_exists(contigs_fasta)

        if self.contigs_fasta is None:
            raise Error('Error finding contigs file: ' + contigs_fasta)

        self.contigs_gfa = self._file_exists(os.path.join(self.assembler_dir, 'contigs.gfa'))
        self.contigs_fastg = self._file_exists(os.path.join(self.assembler_dir, 'contigs.fastg'))
        self.contigs_paths = self._file_exists(os.path.join(self.assembler_dir, 'contigs.paths'))
        self.assembly_graph_fastg = self._file_exists(os.path.join(self.assembler_dir, 'assembly_graph.fastg'))

        if self.assembler == 'spades':
            if None == self.contigs_fastg == self.contigs_paths == self.assembly_graph_fastg or \
               ( self.contigs_fastg is None and (None in {self.contigs_paths, self.assembly_graph_fastg}) ) or \
               ( self.contigs_fastg is not None and (self.contigs_paths is not None or self.assembly_graph_fastg is not None) ):
                error_message = '\n'.join([
                     'Error finding SPAdes graph files in the directory ' + self.assembler_dir,
                     'Expected either:',
                     '    contigs.fastg (SPAdes <3.6.1)',
                     'or:',
                     '    contigs.paths and assembly_graph.fastg (SPAdes >3.6.1)'
                ])
                raise Error(error_message)
        elif self.assembler == 'canu':
            if self.contigs_fasta is None or self.contigs_gfa is None:
                raise Error('Error finding canu contigs fasta and/or gfa file')
        else:
            raise Error('Assembler "' + self.assembler + '" not recognised. Cannot continue')


    def get_contigs(self):
        '''Returns a dictionary of contig_name -> pyfastaq.Sequences.Fasta object'''
        contigs = {}
        pyfastaq.tasks.file_to_dict(self.contigs_fasta, contigs)
        return contigs


    @classmethod
    def _circular_contigs_from_spades_before_3_6_1(cls, fastg_file):
        seq_reader = pyfastaq.sequences.file_reader(fastg_file)
        names = set([x.id.rstrip(';') for x in seq_reader if ':' in x.id])
        found_fwd = set()
        found_rev = set()
        for name in names:
            l = name.split(':')
            if len(l) != 2:
                continue
            if l[0] == l[1]:
                if l[0][-1] == "'":
                    found_rev.add(l[0][:-1])
                else:
                    found_fwd.add(l[0])

        return found_fwd.intersection(found_rev)


    @classmethod
    def _spades_contigs_paths_to_dict(cls, filename):
        d = {}
        node = None

        with open(filename) as f:
            for line in f:
                if node is None:
                    if not line.startswith('NODE_'):
                        raise Error('Error loading info from SPAdes contigs path file ' + filename)
                    node = line.rstrip()
                else:
                    if line.startswith('NODE_') or node in d:
                        raise Error('Error loading info from SPAdes contigs path file ' + filename)
                    d[node] = line.rstrip()
                    node = None

        return d


    @classmethod
    def _circular_edges_to_edge_numbers_dict(cls, circular_edges):
        '''Edge names are like this: EDGE_1_length_44766_cov_16.555.
           This returns dict of edge number -> edge name.
           eg 1 -> EDGE_1_length_44766_cov_16.555.'''
        return {x.split('_')[1]: x for x in circular_edges}


    @staticmethod
    def _circular_contigs_from_spades_after_3_6_1(assembly_graph_fastg, contigs_paths):
        circular_graph_edges = Assembly._circular_contigs_from_spades_before_3_6_1(assembly_graph_fastg)
        circular_edge_dict = Assembly._circular_edges_to_edge_numbers_dict(circular_graph_edges)
        paths_dict = Assembly._spades_contigs_paths_to_dict(contigs_paths)
        circular_nodes = set()

        for node in paths_dict:
            if node.endswith("'"):
                continue

            edges = paths_dict[node].split(',')
            rev_node = node + "'"

            if len(edges) != 1 or rev_node not in paths_dict:
                continue

            rev_edges = paths_dict[rev_node].split(',')

            if len(rev_edges) != 1:
                continue

            edge = list(edges)[0][:-1]
            edge_strand = list(edges)[0][-1]
            rev_edge = list(rev_edges)[0][:-1]
            rev_edge_strand = list(rev_edges)[0][-1]
            if {'-', '+'} == {edge_strand, rev_edge_strand} and edge == rev_edge and edge in circular_edge_dict:
                circular_nodes.add(node)

        return circular_nodes


    @staticmethod
    def _circular_contigs_from_canu_gfa(gfa_file):
        self_matches = {}
        other_matches = set()

        with open(gfa_file) as f:
            for line in f:
                if line.startswith('L\t'):
                    L, node1, dir1, node2, dir2, *the_rest = line.rstrip().split('\t')
                    if node1 == node2:
                        if dir1 == dir2:
                            if node1 not in self_matches:
                                self_matches[node1] = set()
                            self_matches[node1].add(dir1)
                    else:
                        other_matches.update({node1, node2})

        return {x for x in self_matches if self_matches[x] == {'+', '-'} and x not in other_matches}


    def circular_contigs(self):
        '''Returns a set of the contig names that are circular'''
        if self.assembler == 'spades':
            if self.contigs_fastg is not None:
                return self._circular_contigs_from_spades_before_3_6_1(self.contigs_fastg)
            elif None not in [self.contigs_paths, self.assembly_graph_fastg]:
                return self._circular_contigs_from_spades_after_3_6_1(self.assembly_graph_fastg, self.contigs_paths)
            else:
                return set()
        elif self.assembler == 'canu':
            return self._circular_contigs_from_canu_gfa(self.contigs_gfa)
        else:
            return set()



import argparse
import sys
from circlator import versions

def run():
    parser = argparse.ArgumentParser(
        description = 'Checks all dependencies are found and are correct versions',
        usage = 'circlator progcheck'
    )
    parser.add_argument('--debug', action='store_true', help='Debug mode with very verbose output')
    options = parser.parse_args()
    versions.get_all_versions(sys.stdout, raise_error=False, debug=options.debug)



import argparse
import circlator

def run():
    parser = argparse.ArgumentParser(
        description = 'Change start point of each sequence in assembly',
        usage = 'circlator fixstart [options] <assembly.fasta> <outprefix>')
    parser.add_argument('--genes_fa', help='FASTA file of genes to search for to use as start point. If this option is not used, a built-in set of dnaA genes is used', metavar='FILENAME')
    parser.add_argument('--ignore', help='Absolute path to file of IDs of contigs to not change', metavar='FILENAME')
    parser.add_argument('--mincluster', type=int, help='The -c|mincluster option of promer. If this option is used, it overrides promer\'s default value', metavar='INT')
    parser.add_argument('--min_id', type=float, help='Minimum percent identity of promer match between contigs and gene(s) to use as start point [%(default)s]', default=70, metavar='FLOAT')
    parser.add_argument('--verbose', action='store_true', help='Be verbose')
    parser.add_argument('assembly_fa', help='Name of input FASTA file', metavar='assembly.fasta')
    parser.add_argument('outprefix', help='Prefix of output files')
    options = parser.parse_args()

    fixer = circlator.start_fixer.StartFixer(
        options.assembly_fa,
        options.outprefix,
        min_percent_identity=options.min_id,
        promer_mincluster=options.mincluster,
        genes_fa=options.genes_fa,
        ignore=options.ignore,
        verbose=options.verbose,
    )
    fixer.run()



import shutil
import os
import re
import subprocess
from distutils.version import LooseVersion
from circlator import common


class Program:
    def __init__(self, name, version_cmd, version_regex, environment_var=None, debug=False):
        self.name = name
        self.debug = debug

        if environment_var is not None and environment_var in os.environ:
            if self.debug:
                print(self.name, '- getting path from environment variable', environment_var, '=', os.environ[environment_var], flush=True)
            self.path = os.environ[environment_var]
        else:
            if self.debug:
                print(self.name, '- not using environment variable', flush=True)
            self.path = name


        self.version_cmd = version_cmd
        self.version_regex = version_regex

        if self.debug:
            print(self.name, '- checking which(' + self.path + ')', flush=True)

        self.from_which = shutil.which(self.path)

        if self.debug:
            print('   ... got: "', self.from_which, '"', sep='', flush=True)

        self._set_version()


    def in_path(self):
        '''Returns true iff it is in the path'''
        return self.from_which is not None


    def _set_version(self):
        self.version = None

        if self.debug:
            print(self.name, '- checking version ...')

        if not self.in_path():
            if self.debug:
                print(' ... not in path so cannot get version', flush=True)
                return

        cmd = self.exe() + ' ' + self.version_cmd
        if self.debug:
            print('Running this command to get version:', cmd)
        cmd_output = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
        cmd_output = common.decode(cmd_output[0]).split('\n')[:-1] + common.decode(cmd_output[1]).split('\n')[:-1]
        if self.debug:
            print('__________ (begin output from ', cmd, ')___________', sep='')
            print(*cmd_output, sep='\n')
            print('__________ (end of output from ', cmd, ')___________', sep='')
            print('Looking far a match to the regex "', self.version_regex.pattern, '" in the above output', sep='', flush=True)
        for line in cmd_output:
            hits = self.version_regex.search(line)
            if hits:
                if self.debug:
                    print('Match to this line:', line)
                    print('Got version:', hits.group(1), flush=True)
                self.version = hits.group(1)
                break
        else:
            if self.debug:
                print('No match found to the regex', flush=True)



    def version_at_least(self, min_version):
        v = self.version
        if v is None:
            return None
        return LooseVersion(v) >= LooseVersion(min_version)


    def version_at_most(self, max_version):
        v = self.version
        if v is None:
            return None
        return LooseVersion(v) <= LooseVersion(max_version)


    def exe(self):
        '''Returns exectuable that can be used in system call'''
        return self.path




import re
import pyfastaq


class Error (Exception): pass


genetic_code = 11
aa_to_dna = {}
for nucleotide, aa in pyfastaq.genetic_codes.codes[genetic_code].items():
    aa_to_dna[aa] = nucleotide


class UniprotDownloader:
    def __init__(self, min_gene_length=333, max_gene_length=500, uniprot_search='dnaa', header_regex='dnaa', header_regex_ignorecase=True):
        self.min_gene_length = min_gene_length
        self.max_gene_length = max_gene_length
        self.uniprot_search = re.sub('\s+', '+', uniprot_search.strip())
        if header_regex_ignorecase:
            self.header_regex = re.compile(header_regex, re.IGNORECASE)
        else:
            self.header_regex = re.compile(header_regex)


    def _get_uniprot_url(self):
        return 'http://www.uniprot.org/uniprot/?sort=score&desc=&compress=no&query=' + self.uniprot_search + '&force=no&format=fasta'


    def _download_from_uniprot(self, outfile):
        pyfastaq.utils.syscall('wget -O ' + outfile + ' "' + self._get_uniprot_url() + '"')


    def _header_to_genus_species(self, header):
        if 'OS=' not in header:
            return None

        pos = header.find('OS=')
        assert pos != -1
        genus_species = header[pos+3:].split(' ')[:2]
        return tuple(genus_species)


    def _header_matches_regex(self, sequence):
        return self.header_regex.search(sequence.id) is not None


    def _check_sequence(self, sequence, seen_genus_species):
        if not(self.min_gene_length <= len(sequence) <= self.max_gene_length):
            return False, 'Too long or short'
        elif not sequence[0] == 'M':
            return False, 'Does not start with M'
        elif not self._header_matches_regex(sequence):
            return False, 'No match to regex in name'
        else:
            genus_species = self._header_to_genus_species(sequence.id)
            if genus_species is None:
                return False, 'Error getting genus species'
            elif genus_species in seen_genus_species:
                return False, 'Duplicate genus species ' + str(genus_species)
            else:
                seen_genus_species.add(genus_species)
                return True, None


    def _append_stop_to_seq(self, sequence):
        if not sequence[-1] == '*':
            sequence.seq += '*'


    def _reverse_translate(self, sequence):
        amino_acids = list(sequence.seq)
        try:
            nucleotides = [aa_to_dna[aa] for aa in amino_acids]
        except:
            raise Error('Error reverse translate')

        sequence.seq = ''.join(nucleotides)


    def run(self, outprefix):
        aa_file = outprefix + '.aa.fa'
        self._download_from_uniprot(aa_file)
        log_fh = pyfastaq.utils.open_file_write(outprefix + '.log')
        fasta_fh = pyfastaq.utils.open_file_write(outprefix + '.nucleotides.fa')
        seen_genus_species = set()

        for seq in pyfastaq.sequences.file_reader(aa_file):
            ok, error = self._check_sequence(seq, seen_genus_species)

            if ok:
                self._append_stop_to_seq(seq)
                try:
                    self._reverse_translate(seq)
                except:
                    print('Failed reverse translation', seq.id, sep='\t', file=log_fh)
                    continue

                print(seq, file=fasta_fh)
            else:
                print(error, seq.id, sep='\t', file=log_fh)

        pyfastaq.utils.close(log_fh)
        pyfastaq.utils.close(fasta_fh)
